# Import modules
import datetime as dt
import numpy as np
import pandas as pd
import sqlalchemy

# Import custom modules
from etl_tools.sql import sql_exec_stmt, sql_read_data, sql_upload_data, sql_copy_data
from etl_tools.aws import dynamodb_read_data, s3_upload_csv
from etl_tools.api import API_request


class ExtractDeleteAndLoad(object):
    """
    Class with routines to read, delete and/or upload data from/to a database or data storage.
    """

    def __init__(
        self,
        config_dict={},
        conn_dict={},
        sqlalchemy_dict={},
    ):
        """
        Class constructor.

        Parameters:

        config_dict : dict. Configuration dictionary with connection and data parameters.
        conn_dict : dict. Connection dictionary with connection parameters.
        sqlalchemy_dict : dict. Dictionary with sqlalchemy data types.
        """

        ## Set class parameters

        # Set configuration parameters
        self.connections_dict = conn_dict
        self.configs_dict = config_dict
        self.sqlalchemy_dtypes = sqlalchemy_dict
        # Set processes names
        processes_list = ["download", "delete", "truncate", "upload"]
        # Set connection parameters
        self.conn_info_dict = {key: {} for key in processes_list}
        self.conn_suff_dict = {key: {} for key in processes_list}
        self.conn_type_dict = {key: {} for key in processes_list}
        # Iterate over processes
        for p_name in processes_list:
            # Skip if process is not in configuration dictionary
            if f"{p_name}_connections_dict" not in self.configs_dict.keys():
                continue
            # Iterate over connections
            for key in self.configs_dict[f"{p_name}_connections_dict"].keys():
                # Get connection suffix
                self.conn_suff_dict[p_name][key] = self.configs_dict[
                    f"{p_name}_connections_dict"
                ][key].split("_")[-1]
                # Get connection type
                self.conn_type_dict[p_name][key] = self.configs_dict[
                    f"{p_name}_connections_dict"
                ][key].split("_")[0]
                # Get connection dictionary
                self.conn_info_dict[p_name][key] = {
                    "myoracle_client_dir": (
                        self.connections_dict["myoracle_client_dir"]
                        if "myoracle_client_dir" in self.connections_dict.keys()
                        else ""
                    ),
                    "myserver": (
                        self.connections_dict[
                            f"myserver_{self.conn_type_dict[p_name][key]}_{self.conn_suff_dict[p_name][key]}"
                        ]
                        if f"myserver_{self.conn_type_dict[p_name][key]}_{self.conn_suff_dict[p_name][key]}"
                        in self.connections_dict.keys()
                        else ""
                    ),
                    "mydatabase": (
                        self.connections_dict[
                            f"mydatabase_{self.conn_type_dict[p_name][key]}_{self.conn_suff_dict[p_name][key]}"
                        ]
                        if f"mydatabase_{self.conn_type_dict[p_name][key]}_{self.conn_suff_dict[p_name][key]}"
                        in self.connections_dict.keys()
                        else ""
                    ),
                    "myusername": (
                        self.connections_dict[
                            f"myusername_{self.conn_type_dict[p_name][key]}_{self.conn_suff_dict[p_name][key]}"
                        ]
                        if f"myusername_{self.conn_type_dict[p_name][key]}_{self.conn_suff_dict[p_name][key]}"
                        in self.connections_dict.keys()
                        else ""
                    ),
                    "mypassword": (
                        self.connections_dict[
                            f"mypassword_{self.conn_type_dict[p_name][key]}_{self.conn_suff_dict[p_name][key]}"
                        ]
                        if f"mypassword_{self.conn_type_dict[p_name][key]}_{self.conn_suff_dict[p_name][key]}"
                        in self.connections_dict.keys()
                        else ""
                    ),
                    "mycharset": (
                        self.connections_dict[
                            f"mycharset_{self.conn_type_dict[p_name][key]}_{self.conn_suff_dict[p_name][key]}"
                        ]
                        if f"mycharset_{self.conn_type_dict[p_name][key]}_{self.conn_suff_dict[p_name][key]}"
                        in self.connections_dict.keys()
                        else ""
                    ),
                    "myencoding": (
                        self.connections_dict[
                            f"awmyencoding_{self.conn_type_dict[p_name][key]}_{self.conn_suff_dict[p_name][key]}"
                        ]
                        if f"awmyencoding_{self.conn_type_dict[p_name][key]}_{self.conn_suff_dict[p_name][key]}"
                        in self.connections_dict.keys()
                        else ""
                    ),
                }

    def delete_data(self, **kwargs):
        """
        Function to delete data from the source.

        Parameters:

        kwargs : dict. Keyword arguments to pass to the delete statement.
        """

        # Check if delete configurations are set
        if (
            ("delete" not in self.conn_suff_dict.keys())
            or ("delete" not in self.conn_type_dict.keys())
            or ("delete" not in self.conn_info_dict.keys())
        ):
            raise ValueError(
                "Delete configurations are not set! Please set them first."
            )
        # Set keyword arguments as global variables
        for key, val in kwargs.items():
            globals()[key] = val
        # Iterate over table/statements
        for key in (
            self.configs_dict["delete_tables_dict"].keys()
            if "delete_tables_dict" in self.configs_dict.keys()
            else self.configs_dict["delete_sql_stmts_dict"].keys()
        ):
            print(f"Deleting data for {key}...")
            # Set extra variables as global variables
            if "delete_extra_vars_dict" in self.configs_dict.keys():
                for ex_key, ex_val in self.configs_dict["delete_extra_vars_dict"][
                    key
                ].items():
                    globals()[ex_key] = eval(ex_val) if "eval(" in ex_val else ex_val
            # Get connection suffix
            conn_suff = self.conn_suff_dict["delete"][key]
            # Get connection type
            conn_type = self.conn_type_dict["delete"][key]
            # Get connection dictionary
            conn_dict = self.conn_info_dict["delete"][key]
            # Delete data
            if conn_type == "dynamodb":
                pass
            elif conn_type == "s3":
                pass
            elif conn_type == "api":
                api_response = API_request(
                    self.configs_dict["delete_tables_dict"][key],
                    headers=(
                        self.configs_dict["delete_headers_dict"][key]
                        if "delete_headers_dict" in self.configs_dict.keys()
                        else None
                    ),
                    params=(
                        self.configs_dict["delete_params_dict"][key]
                        if "delete_params_dict" in self.configs_dict.keys()
                        else None
                    ),
                    data=(
                        self.configs_dict["delete_datas_dict"][key]
                        if "delete_datas_dict" in self.configs_dict.keys()
                        and key in self.configs_dict["delete_datas_dict"].keys()
                        else None
                    ),
                    json=(
                        self.configs_dict["delete_jsons_dict"][key]
                        if "delete_jsons_dict" in self.configs_dict.keys()
                        and key in self.configs_dict["delete_jsons_dict"].keys()
                        else None
                    ),
                    request_type=(self.configs_dict["delete_request_types_dict"][key]),
                )
                print(f"     API response: {api_response}")
            else:
                # Get delete statement
                stmt = (
                    eval(self.configs_dict["delete_sql_stmts_dict"][key])
                    if "{" in self.configs_dict["delete_sql_stmts_dict"][key]
                    else self.configs_dict["delete_sql_stmts_dict"][key]
                )
                print(f"     Delete query: {stmt}")
                # Execute delete statement
                try:
                    sql_exec_stmt(
                        stmt,
                        conn_dict,
                        mode=conn_type,
                    )
                except Exception as e:
                    print(f"Error deleting data: {type(e)} - {e}")

        pass

    def truncate_data(self, **kwargs):
        """
        Function to truncate data from the source.
        """

        # Check if truncate configurations are set
        if (
            ("truncate" not in self.conn_suff_dict.keys())
            or ("truncate" not in self.conn_type_dict.keys())
            or ("truncate" not in self.conn_info_dict.keys())
        ):
            raise ValueError(
                "Truncate configurations are not set! Please set them first."
            )
        # Set keyword arguments as global variables
        for key, val in kwargs.items():
            globals()[key] = val
        # Iterate over tables/statements
        for key in (
            self.configs_dict["truncate_tables_dict"].keys()
            if "truncate_tables_dict" in self.configs_dict.keys()
            else self.configs_dict["truncate_sql_stmts_dict"].keys()
        ):
            print(f"Truncating data for {key}...")
            # Get connection suffix
            conn_suff = self.conn_suff_dict["truncate"][key]
            # Get connection type
            conn_type = self.conn_type_dict["truncate"][key]
            # Get connection dictionary
            conn_dict = self.conn_info_dict["truncate"][key]
            # Truncate data
            if conn_type == "dynamodb":
                pass
            elif conn_type == "s3":
                pass
            elif conn_type == "api":
                pass
            else:
                # Get truncate statement
                stmt = (
                    eval(self.configs_dict["truncate_sql_stmts_dict"][key])
                    if "{" in self.configs_dict["truncate_sql_stmts_dict"][key]
                    else self.configs_dict["truncate_sql_stmts_dict"][key]
                )
                print(f"     Truncate query: {stmt}")
                # Execute truncate statement
                try:
                    sql_exec_stmt(
                        stmt,
                        conn_dict,
                        mode=conn_type,
                    )
                except Exception as e:
                    print(f"Error truncating data: {type(e)} - {e}")

        pass

    def read_data(self, **kwargs):
        """
        Function to read data from the source.
        """

        # Check if download configurations are set
        if (
            ("download" not in self.conn_suff_dict.keys())
            or ("download" not in self.conn_type_dict.keys())
            or ("download" not in self.conn_info_dict.keys())
        ):
            raise ValueError(
                "Download configurations are not set! Please set them first."
            )
        # Set keyword arguments as global variables
        for key, val in kwargs.items():
            globals()[key] = val
        # Initialize raw data
        self.raw_data = {}
        # Iterate over tables/statements
        for key in (
            self.configs_dict["download_tables_dict"].keys()
            if "download_tables_dict" in self.configs_dict.keys()
            else self.configs_dict["download_sql_stmts_dict"].keys()
        ):
            print(f"Downloading data for {key}...")
            # Set extra variables as global variables
            if "download_extra_vars_dict" in self.configs_dict.keys():
                for ex_key, ex_val in self.configs_dict["download_extra_vars_dict"][
                    key
                ].items():
                    globals()[ex_key] = eval(ex_val) if "eval(" in ex_val else ex_val
            # Get connection suffix
            conn_suff = self.conn_suff_dict["download"][key]
            # Get connection type
            conn_type = self.conn_type_dict["download"][key]
            # Get connection dictionary
            conn_dict = self.conn_info_dict["download"][key]
            # Read data
            if conn_type == "dynamodb":
                print(
                    f"     DynamoDB table: {self.configs_dict['download_tables_dict'][key]}"
                )
                # Evaluate keyword arguments
                dynamo_kwargs_eval = {
                    kw_key: eval(kw_val) if "{" in kw_val else kw_val
                    for kw_key, kw_val in self.configs_dict[
                        "download_dynamodb_kwargs_dict"
                    ][key].items()
                }
                # Download data from DynamoDB
                data = dynamodb_read_data(
                    self.configs_dict["download_tables_dict"][key],
                    self.connections_dict[f"aws_access_key_id_{conn_suff}"],
                    self.connections_dict[f"aws_secret_access_key_{conn_suff}"],
                    self.connections_dict[f"region_name_{conn_suff}"],
                    **dynamo_kwargs_eval,
                )
            elif conn_type == "s3":
                pass
            elif conn_type == "api":
                print(
                    f"     API endpoint: {self.configs_dict['download_tables_dict'][key]}"
                )
                # Download data from API
                data = API_request(
                    self.configs_dict["download_tables_dict"][key],
                    headers=(
                        self.configs_dict["download_headers_dict"][key]
                        if "download_headers_dict" in self.configs_dict.keys()
                        else None
                    ),
                    params=(
                        self.configs_dict["download_params_dict"][key]
                        if "download_params_dict" in self.configs_dict.keys()
                        else None
                    ),
                    data=(
                        self.configs_dict["download_datas_dict"][key]
                        if "download_datas_dict" in self.configs_dict.keys()
                        and key in self.configs_dict["download_datas_dict"].keys()
                        else None
                    ),
                    json=(
                        self.configs_dict["download_jsons_dict"][key]
                        if "download_jsons_dict" in self.configs_dict.keys()
                        and key in self.configs_dict["download_jsons_dict"].keys()
                        else None
                    ),
                    request_type=(
                        self.configs_dict["download_request_types_dict"][key]
                    ),
                )
            else:
                # Get download statement
                stmt = (
                    eval(self.configs_dict["download_sql_stmts_dict"][key])
                    if "{" in self.configs_dict["download_sql_stmts_dict"][key]
                    else self.configs_dict["download_sql_stmts_dict"][key]
                )
                print(f"     Download query: {stmt}")
                # Download data
                data = sql_read_data(
                    stmt,
                    conn_dict,
                    mode=conn_type,
                    custom_conn_str=(
                        self.configs_dict["download_custom_conn_strs_dict"][key]
                        if "download_custom_conn_strs_dict" in self.configs_dict.keys()
                        else None
                    ),
                    connect_args=(
                        self.configs_dict["download_connect_args_dict"][key]
                        if "download_connect_args_dict" in self.configs_dict.keys()
                        else {}
                    ),
                    name=(
                        self.configs_dict["download_tables_dict"][key]
                        if "download_tables_dict" in self.configs_dict.keys()
                        else key
                    ),
                    max_n_try=(
                        self.configs_dict["max_n_try"]
                        if "max_n_try" in self.configs_dict.keys()
                        else 3
                    ),
                )
            # Add data to raw data dictionary
            self.raw_data[key] = data.copy()
        # Free memory
        del data

        pass

    def upload_data(self, data_to_upload: dict, **kwargs):
        """
        Function to upload data to the target.

        Parameters:

        data_to_upload : list. List with data to upload.
        """

        # Check if upload configurations are set
        if (
            ("upload" not in self.conn_suff_dict.keys())
            or ("upload" not in self.conn_type_dict.keys())
            or ("upload" not in self.conn_info_dict.keys())
        ):
            raise ValueError(
                "Upload configurations are not set! Please set them first."
            )
        # Set keyword arguments as global variables
        for key, val in kwargs.items():
            globals()[key] = val
        # Iterate over tables/statements
        for key in (
            self.configs_dict["upload_tables_dict"].keys()
            if "upload_tables_dict" in self.configs_dict.keys()
            else self.configs_dict["upload_sql_stmts_dict"].keys()
        ):
            print(f"Uploading data for {key}...")
            # Set data to upload
            upload_data = data_to_upload[key]
            # Get connection suffix
            conn_suff = self.conn_suff_dict["upload"][key]
            # Get connection type
            conn_type = self.conn_type_dict["upload"][key]
            # Get connection dictionary
            conn_dict = self.conn_info_dict["upload"][key]
            # Upload data
            if conn_type == "dynamodb":
                print(
                    f"     DynamoDB table: {self.configs_dict['upload_tables_dict'][key]}"
                )
                # Upload data to DynamoDB
                pass
            elif conn_type == "s3":
                print(f"     S3 bucket: {self.configs_dict['upload_tables_dict'][key]}")
                # Upload data to S3
                s3_upload_csv(
                    upload_data,
                    self.configs_dict["s3_file_paths_dict"][key],
                    self.connections_dict[f"aws_access_key_id_{conn_suff}"],
                    self.connections_dict[f"aws_secret_access_key_{conn_suff}"],
                    region_name=self.connections_dict[f"region_name_{conn_suff}"],
                    sep=(
                        self.configs_dict["s3_csv_seps_dict"][key]
                        if "s3_csv_seps_dict" in self.configs_dict.keys()
                        else ","
                    ),
                    index=False,
                    encoding=(
                        self.configs_dict["s3_csv_encodings_dict"][key]
                        if "s3_csv_encodings_dict" in self.configs_dict.keys()
                        else "utf-8"
                    ),
                )
            elif conn_type == "api":
                print(
                    f"     API endpoint: {self.configs_dict['upload_tables_dict'][key]}"
                )
                # Upload data to API
                api_response = API_request(
                    self.configs_dict["upload_tables_dict"][key],
                    headers=(
                        self.configs_dict["upload_headers_dict"][key]
                        if "upload_headers_dict" in self.configs_dict.keys()
                        else None
                    ),
                    data=(
                        upload_data
                        if "upload_datas_dict" in self.configs_dict.keys()
                        and key in self.configs_dict["upload_datas_dict"].keys()
                        else None
                    ),
                    json=(
                        upload_data
                        if "upload_jsons_dict" in self.configs_dict.keys()
                        and key in self.configs_dict["upload_jsons_dict"].keys()
                        else None
                    ),
                    request_type=self.configs_dict["upload_request_types_dict"][key],
                )
                print(f"     API response: {api_response}")
            else:
                # Define column names and data types
                col_dict = self.configs_dict["upload_python_to_sql_dtypes_dict"][key]
                # Define dictionary with data types
                dtypes_dict = {
                    col: (
                        eval(
                            f'{self.sqlalchemy_dtypes[col_dtype.split("(")[0]]}({col_dtype.split("(")[1][:-1]})'
                        )
                        if "(" in col_dtype
                        else eval(f"{self.sqlalchemy_dtypes[col_dtype]}()")
                    )
                    for col, col_dtype in col_dict.items()
                }
                # Use order defined in data types dictionary
                upload_data = upload_data[list(col_dict.keys())]
                # Upload data to created tables
                if "redshift" in conn_type:  # Upload data to Redshift
                    print(
                        f"     Redshift table: {self.configs_dict['upload_tables_dict'][key]}"
                    )
                    # Upload data to S3
                    s3_upload_csv(
                        upload_data,
                        self.configs_dict["s3_file_paths_dict"][key],
                        self.connections_dict[f"aws_access_key_id_{conn_suff}"],
                        self.connections_dict[f"aws_secret_access_key_{conn_suff}"],
                        region_name=self.connections_dict[f"region_name_{conn_suff}"],
                        sep=(
                            self.configs_dict["s3_csv_seps_dict"][key]
                            if "s3_csv_seps_dict" in self.configs_dict.keys()
                            else ","
                        ),
                        index=False,
                        encoding=(
                            self.configs_dict["s3_csv_encodings_dict"][key]
                            if "s3_csv_encodings_dict" in self.configs_dict.keys()
                            else "utf-8"
                        ),
                    )
                    # Copy data from S3 to database
                    sql_copy_data(
                        (
                            eval(self.configs_dict["s3_file_paths_dict"][key])
                            if "{" in self.configs_dict["s3_file_paths_dict"][key]
                            else self.configs_dict["s3_file_paths_dict"][key]
                        ),
                        self.configs_dict["upload_schemas_dict"][key],
                        self.configs_dict["upload_tables_dict"][key],
                        conn_dict,
                        self.connections_dict[f"aws_access_key_id_aws_{conn_suff}"],
                        self.connections_dict[f"aws_secret_access_key_aws_{conn_suff}"],
                        self.connections_dict[f"region_name_aws_{conn_suff}"],
                        name=self.configs_dict["upload_tables_dict"][key],
                        max_n_try=(
                            self.configs_dict["max_n_try"]
                            if "max_n_try" in self.configs_dict.keys()
                            else 3
                        ),
                    )
                else:  # Upload data to other databases
                    print(
                        f"     {conn_type.capitalize()} table: {self.configs_dict['upload_tables_dict'][key]}"
                    )
                    # Upload data to database
                    sql_upload_data(
                        upload_data,
                        self.configs_dict["upload_schemas_dict"][key],
                        self.configs_dict["upload_tables_dict"][key],
                        conn_dict,
                        mode=conn_type,
                        custom_conn_str=(
                            self.configs_dict["upload_custom_conn_strs_dict"][key]
                            if "upload_custom_conn_strs_dict"
                            in self.configs_dict.keys()
                            else None
                        ),
                        connect_args=(
                            self.configs_dict["upload_connect_args_dict"][key]
                            if "upload_connect_args_dict" in self.configs_dict.keys()
                            else {}
                        ),
                        name=self.configs_dict["upload_tables_dict"][key],
                        chunksize=(
                            self.configs_dict["upload_chunksizes_dict"][key]
                            if "upload_chunksizes_dict" in self.configs_dict.keys()
                            else 100
                        ),
                        method=(
                            self.configs_dict["upload_methods_dict"][key]
                            if "upload_methods_dict" in self.configs_dict.keys()
                            else "multi"
                        ),
                        dtypes_dict=dtypes_dict,
                        max_n_try=(
                            self.configs_dict["max_n_try"]
                            if "max_n_try" in self.configs_dict.keys()
                            else 3
                        ),
                        n_jobs=(
                            self.configs_dict["n_parallel"]
                            if "n_parallel" in self.configs_dict.keys()
                            else -1
                        ),
                    )

        pass



# Import modules
import os
import sys
import traceback
import pyspark as ps
import pyodbc
import redshift_connector
import oracledb
import numpy as np
import pandas as pd
import datetime as dt
import multiprocessing

# Import submodules
from sqlalchemy import create_engine

# Import custom modules
from etl_tools.execution import mk_err_logs, mk_texec_logs, parallel_execute


def create_sqlalchemy_engine(
    conn_dict: dict,
    custom_conn_str=None,
    connect_args={},
):
    """
    Function to create sqlalchemy enginefrom connection dictionary.

    conn_dict :                 Dictionary with server, database, uid and pwd information.
    custom_conn_str :           String with custom connection string.
    connect_args :              Dictionary with extra arguments for connection.
    """

    if custom_conn_str:
        ## Create engine
        engine = create_engine(custom_conn_str, connect_args=connect_args)
    else:
        ## Set extra configuration for connection

        # Engine prefix
        if "myengine_prefix" not in conn_dict.keys():
            conn_dict["myengine_prefix"] = "mssql+pyodbc"
        # Port
        if "myport" not in conn_dict.keys():
            conn_dict["myport"] = 1433

        ## Create engine

        engine = create_engine(
            f'{conn_dict["myengine_prefix"]}://{conn_dict["myusername"]}:{conn_dict["mypassword"]}@{conn_dict["myserver"]}:{conn_dict["myport"]}/{conn_dict["mydatabase"]}'
        )

    return engine


def create_redshift_engine(conn_dict: dict):
    """
    Function to create redshift engine from connection dictionary.
    """

    ## Set extra configuration for connection

    # Port
    if "myport" not in conn_dict.keys():
        conn_dict["myport"] = 5439
    # SSL mode
    if "mysslmode" not in conn_dict.keys():
        conn_dict["mysslmode"] = "verify-ca"
    # Connect args
    connect_args = {"sslmode": conn_dict["mysslmode"]}

    ## Create custom connection string
    custom_conn_str = f'redshift+redshift_connector://{conn_dict["myusername"]}:{conn_dict["mypassword"]}@{conn_dict["myserver"]}:{conn_dict["myport"]}/{conn_dict["mydatabase"]}'

    ## Create engine
    engine = create_sqlalchemy_engine(
        conn_dict, custom_conn_str=custom_conn_str, connect_args=connect_args
    )

    return engine


def create_oracle_engine(conn_dict: dict):
    """
    Function to create oracle engine from connection dictionary.
    """

    ## Set extra configuration for connection

    ## Start oracle client
    try:
        oracledb.init_oracle_client(lib_dir=conn_dict["myoracle_client_dir"])  # Windows
    except:
        try:
            oracledb.init_oracle_client()  # Linux
        except:
            pass  # Oracle client already started or not needed
    # Port
    if "myport" not in conn_dict.keys():
        conn_dict["myport"] = 1521

    ## Create custom connection string
    custom_conn_str = f'oracle+cx_oracle://{conn_dict["myusername"]}:{conn_dict["mypassword"]}@{conn_dict["myserver"]}:{conn_dict["myport"]}/?service_name={conn_dict["mydatabase"]}'

    ## Create engine
    engine = create_sqlalchemy_engine(conn_dict, custom_conn_str=custom_conn_str)

    return engine


def create_mysql_engine(conn_dict: dict):
    """
    Function to create mysql engine from connection dictionary.
    """

    ## Set extra configuration for connection

    # Port
    if "myport" not in conn_dict.keys():
        conn_dict["myport"] = 3306
    # Charset
    if "mycharset" not in conn_dict.keys():
        conn_dict["mycharset"] = "utf8mb4"
    # Connect args
    connect_args = {
        "charset": conn_dict["mycharset"],
    }

    ## Create custom connection string
    custom_conn_str = f'mysql+pymysql://{conn_dict["myusername"]}:{conn_dict["mypassword"]}@{conn_dict["myserver"]}:{conn_dict["myport"]}/{conn_dict["mydatabase"]}'

    ## Create engine
    engine = create_sqlalchemy_engine(
        conn_dict,
        custom_conn_str=custom_conn_str,
        connect_args=connect_args,
    )

    return engine


def create_sqlalchemy_conn(conn_dict: dict, custom_conn_str=None):
    """
    Function to create sqlalchemy connector from connection dictionary.
    """
    ## Create engine
    engine = create_sqlalchemy_engine(conn_dict, custom_conn_str=custom_conn_str)

    print("Connecting to database...")
    ## Make connection
    conn = engine.connect()

    return conn


def create_redshift_conn(conn_dict: dict):
    """
    Function to create redshift connector from connection dictionary.
    """

    ## Set extra configuration for connection

    # Port
    if "myport" not in conn_dict.keys():
        conn_dict["myport"] = 5439

    ## Create connector

    conn = redshift_connector.connect(
        host=conn_dict["myserver"],
        database=conn_dict["mydatabase"],
        port=conn_dict["myport"],
        user=conn_dict["myusername"],
        password=conn_dict["mypassword"],
    )

    return conn


def create_oracle_conn(conn_dict: dict):
    """
    Function to create oracle connector from connection dictionary.
    """

    ## Start oracle client
    try:
        oracledb.init_oracle_client(lib_dir=conn_dict["myoracle_client_dir"])  # Windows
    except:
        try:
            oracledb.init_oracle_client()  # Linux
        except:
            pass  # Oracle client already started or not needed

    ## Create connector
    conn = oracledb.connect(
        user=conn_dict["myusername"],
        password=conn_dict["mypassword"],
        dsn=f'{conn_dict["myserver"]}/{conn_dict["mydatabase"]}',
    )

    return conn


def create_pyodbc_conn(conn_dict: dict):
    """
    Function to create pyodbc connector from connection dictionary.
    """

    ## Set driver for connection

    if "mydriver" not in conn_dict.keys():
        conn_dict["mydriver"] = "{ODBC Driver 17 for SQL Server}"  # '{SQL Server}'

    ## Create string for connection

    str_conn = (
        f'DRIVER={conn_dict["mydriver"]};'
        + f'SERVER={conn_dict["myserver"]};'
        + f'DATABASE={conn_dict["mydatabase"]};'
        + f'UID={conn_dict["myusername"]};'
        + f'PWD={conn_dict["mypassword"]}'
    )

    ## Create connector
    conn = pyodbc.connect(str_conn)

    return conn


def sql_exec_stmt(sql_stmt, conn_dict: dict, mode="pyodbc"):
    """
    Function to execute sql statements.

    sql_stmt :                  String with sql statement to execute.
    conn_dict :                 Dictionary with server, database, uid and pwd information.
    mode :                      String with mode to use. Options are 'pyodbc' and 'redshift'.
    """

    ## Set mode of connection

    if mode == "pyodbc":
        print("Connecting to database...")
        ## Make connection
        sql_conn = create_pyodbc_conn(conn_dict)

    elif mode == "redshift":
        print("Connecting to database...")
        ## Make connection
        sql_conn = create_redshift_conn(conn_dict)

    elif mode == "sqlalchemy":
        print("Connecting to database...")
        ## Make connection
        sql_conn = create_sqlalchemy_conn(conn_dict)

    elif mode == "oracledb":
        print("Connecting to database...")
        ## Make connection
        sql_conn = create_oracle_conn(conn_dict)

    print("Executing statement...")
    ## Execute statment

    with sql_conn:
        # Initialize cursor
        cursor = sql_conn.cursor()
        # Execute statement
        cursor.execute(sql_stmt)
        # Get number of rows affected
        response_rows_affected = cursor.rowcount
        # Commit changes
        sql_conn.commit()

    return response_rows_affected  # Response with number of rows affected


def to_sql_executemany(
    data,
    conn_dict,
    schema,
    table_name,
    mode,
):
    """
    Function to upload data to database table with sqlalchemy in parallel.

    data :                      Pandas dataframe with data to upload.
    conn_dict :                 Dictionary with server, database, uid and pwd information.
    schema :                    String with schema name.
    table_name :                String with table name to upload data.
    mode :                      String with mode to use. Options are 'pyodbc' and 'redshift'.
    """

    ## Set mode of connection

    if mode == "pyodbc":
        print("Connecting to database...")
        ## Make connection
        sql_conn = create_pyodbc_conn(conn_dict)

    elif mode == "redshift":
        print("Connecting to database...")
        ## Make connection
        sql_conn = create_redshift_conn(conn_dict)

    elif mode == "sqlalchemy":
        print("Connecting to database...")
        ## Make connection
        sql_conn = create_sqlalchemy_conn(conn_dict)

    elif mode == "oracledb":
        print("Connecting to database...")
        ## Make connection
        sql_conn = create_oracle_conn(conn_dict)

    print("Executing statement...")
    ## Execute statment

    # Create sql statement
    sql_stmt = f"INSERT INTO {schema}.{table_name} ({','.join([col for col in data.columns])}) VALUES ({','.join([f':{col}' for col in data.columns])})"
    # Format data into sequence
    data = [tuple(x) for x in data.to_numpy()]
    # Make query
    with sql_conn:
        # Initialize cursor
        cursor = sql_conn.cursor()
        # Execute statement
        cursor.executemany(sql_stmt, data)
        # Commit changes
        sql_conn.commit()

    return len(
        data
    )  # Response with number of rows inserted (should be equal to len(data))


def to_sql_redshift_spark(data, schema, table_name, conn_dict, mode="append"):
    """
    Function to upload data to redshift with spark.

    data :                      Pandas dataframe with data to upload.
    schema :                    String with schema name.
    table_name :                String with table name to upload data.
    conn_dict :                 Dictionary with server, database, uid and pwd information.
    mode :                      String with mode to use. Options are 'append', 'overwrite', 'ignore' and 'error'.
    """

    # Create spark session
    spark_session = (
        ps.sql.SparkSession.builder.appName("UploadDataPipeline")
        .enableHiveSupport()
        .getOrCreate()
    )
    # Create spark dataframe
    spark_df = spark_session.createDataFrame(data)
    # Upload data to redshift (options configuration from https://spark.apache.org/docs/latest/sql-data-sources-jdbc.html#data-source-option)
    spark_df.write.format("io.github.spark_redshift_community.spark.redshift").option(
        "url",
        f'jdbc:redshift://{conn_dict["mydatabase"]}',
    ).option("dbtable", f"{schema}.{table_name}").option(
        "user", conn_dict["myusername"]
    ).option(
        "password", conn_dict["mypassword"]
    ).option(
        "tempdir", "s3://tmp-spark/"
    ).mode(
        mode
    ).save()
    # Get response with number of rows inserted
    response_rows_affected = spark_df.count()

    return response_rows_affected


def parallel_to_sql(
    df,
    table_name,
    schema,
    mode,
    conn_dict,
    custom_conn_str,
    connect_args,
    chunksize,
    method,
    dtypes_dict,
    spark_mode="append",
):
    """
    Function to upload data to database table with sqlalchemy in parallel.

    df :                        Pandas dataframe with data to upload.
    table_name :                String with table name to upload data.
    engine :                    SQLAlchemy engine.
    schema :                    String with schema name.
    mode :                      String with mode to use. Options are 'sqlalchemy', 'redshift' and 'oracledb.
    conn_dict :                 Dictionary with server, database, uid and pwd information.
    custom_conn_str :           String with custom connection string.
    connect_args :              Dictionary with extra arguments for connection.
    chunksize :                 Integer with chunksize to use.
    method :                    String with method to use ('multi', 'execute_many', 'spark' or 'single').
    dtypes_dict :               Dictionary with dtypes to use for upload.
    spark_mode :                String with mode to use when uploading to redshift with spark. Options are 'append', 'overwrite', 'ignore' and 'error'.
    """

    print("Connecting to database...")
    ## Create engine connection

    if mode.lower() == "sqlalchemy":
        engine = create_sqlalchemy_engine(
            conn_dict,
            custom_conn_str=custom_conn_str,
            connect_args=connect_args,
        )
    elif mode.lower() == "redshift":
        engine = create_redshift_engine(conn_dict)
    elif mode.lower() == "oracledb":
        engine = create_oracle_engine(conn_dict)

    print("Uploading data...")
    ## Upload data

    # Initialize response rows affected variable
    tot_response_rows_affected = 0
    # Upload data to database
    if method.lower() == "multi":
        try:
            print("Trying to upload data with 'multi' method...")
            response_rows_affected = df.to_sql(
                table_name,
                engine,
                schema=schema,
                if_exists="append",
                index=False,
                chunksize=chunksize,
                method=method,
                dtype=dtypes_dict,
            )
        except Exception as e:
            print(f"{type(e)} - {e}")
            try:
                print(
                    "Uploading data with 'multi' method failed. Trying to upload data with 'execute many' method..."
                )
                response_rows_affected = to_sql_executemany(
                    df, conn_dict, schema, table_name, mode
                )
            except Exception as e:
                print(f"{type(e)} - {e}")
                print(
                    "Uploading data with 'execute many' method failed. Trying to upload data with 'single' method..."
                )
                response_rows_affected = df.to_sql(
                    table_name,
                    engine,
                    schema=schema,
                    if_exists="append",
                    index=False,
                    chunksize=chunksize,
                    dtype=dtypes_dict,
                )
    elif method.lower() == "execute_many":
        try:
            print("Trying to upload data with 'execute many' method...")
            response_rows_affected = to_sql_executemany(
                df, conn_dict, schema, table_name, mode
            )
        except Exception as e:
            print(f"{type(e)} - {e}")
            print(
                "Uploading data with 'execute many' method failed. Trying to upload data with 'single' method..."
            )
            response_rows_affected = df.to_sql(
                table_name,
                engine,
                schema=schema,
                if_exists="append",
                index=False,
                chunksize=chunksize,
                dtype=dtypes_dict,
            )
    elif method.lower() == "spark":
        try:
            print("Trying to upload data with 'spark' method...")
            response_rows_affected = to_sql_redshift_spark(
                df, schema, table_name, conn_dict, mode=spark_mode
            )
        except Exception as e:
            print(f"{type(e)} - {e}")
            print(
                "Uploading data with 'spark' method failed. Trying to upload data with 'single' method..."
            )
            response_rows_affected = df.to_sql(
                table_name,
                engine,
                schema=schema,
                if_exists="append",
                index=False,
                chunksize=chunksize,
                dtype=dtypes_dict,
            )
    elif method.lower() == "single":
        print("Trying to upload data with 'single' method...")
        response_rows_affected = df.to_sql(
            table_name,
            engine,
            schema=schema,
            if_exists="append",
            index=False,
            chunksize=chunksize,
            dtype=dtypes_dict,
        )
    else:
        print("Wrong selected method. Aborting...")
        response_rows_affected = 0

    # Update response rows affected
    tot_response_rows_affected += response_rows_affected

    return tot_response_rows_affected


def sql_read_data(
    sql_stmt,
    conn_dict,
    mode="sqlalchemy",
    custom_conn_str=None,
    connect_args={},
    name=None,
    max_n_try=3,
):
    """
    Function to read sql statements.

    sql_stmt :            SQL statement to execute.
    conn_dict :            Dictionary with server, database, uid and pwd information.
    mode :                 Mode to use. Options are 'sqlalchemy', 'redshift' and 'oracledb'.
    custom_conn_str :      Custom connection string.
    connect_args :         Custom connection argument.
    name :                 Name to use for print statements.
    max_n_try :            Maximum number of tries to execute the query.
    """

    ## Read data

    t_i = dt.datetime.now()
    n_try = 0
    succeeded = False
    while n_try < max_n_try and not succeeded:
        try:
            # Create engine
            if mode == "oracledb":
                engine_obj = create_oracle_engine(conn_dict)
            elif mode == "redshift":
                engine_obj = create_redshift_engine(conn_dict)
            elif mode == "sqlalchemy":
                engine_obj = create_sqlalchemy_engine(
                    conn_dict,
                    custom_conn_str=custom_conn_str,
                    connect_args=connect_args,
                )
            # Dispose connections
            engine_obj.dispose()
            # Read sql statement
            df = pd.read_sql(sql_stmt, engine_obj)
            # Change status
            succeeded = True
        except Exception as e:
            # Initialize empty dataframe
            df = pd.DataFrame()
            # Set log path
            log_file_path = "logs"
            # Set log file name
            log_file_name = "read_data"
            # Create logs folder
            os.makedirs(log_file_path, exist_ok=True)
            # Create logs
            mk_err_logs(
                log_file_path,
                log_file_name,
                sys._getframe().f_code.co_name + " -> " + name,
                traceback.format_exception(
                    sys.exc_info()[0], sys.exc_info()[1], sys.exc_info()[2]
                )[-1],
                mode="summary",
            )
            mk_err_logs(
                log_file_path,
                log_file_name,
                sys._getframe().f_code.co_name + " -> " + name,
                "".join(
                    traceback.format_exception(
                        sys.exc_info()[0], sys.exc_info()[1], sys.exc_info()[2]
                    )
                ),
                mode="detailed",
            )
            # Change status
            succeeded = False
        # Update number of tries
        n_try += 1
    t_e = dt.datetime.now()
    print(f"Time elapsed in 'download' process {name} -> {df.shape} = {t_e - t_i}")

    ## Create time execution logs

    # Set log path
    log_file_path = "logs"
    # Set log file name
    log_file_name = "download_data_texec"
    # Create logs folder
    os.makedirs(log_file_path, exist_ok=True)
    # Create logs
    mk_texec_logs(
        log_file_path,
        log_file_name,
        sys._getframe().f_code.co_name + " -> " + name,
        t_e - t_i,
        obs=f"Shape of object = {df.shape}",
    )

    return df


def sql_upload_data(
    df,
    schema,
    table_name,
    conn_dict,
    mode="sqlalchemy",
    custom_conn_str=None,
    connect_args={},
    name=None,
    chunksize=1000,
    method="multi",
    max_n_try=3,
    dtypes_dict={},
    n_jobs=-1,
    spark_mode="append",
):
    """
    Function to upload data to database table with sqlalchemy.

    df :                    Dataframe to upload.
    schema :                Schema to upload data to.
    table_name :            Table name to upload data to.
    conn_dict :             Dictionarie with server, database, uid and pwd information.
    mode :                  String with mode to use. Options are 'sqlalchemy' and 'redshift'.
    custom_conn_str :       String with custom connection string.
    connect_args :          Dictionarie with connection arguments.
    name :                  Name to use for print statements.
    chunksize :             Integer with chunksize to use for upload.
    method :                String with method to use for upload ('multi', 'execute_many' or 'single').
    max_n_try :             Integer with maximum number of tries to upload data.
    dtypes_dict :           Dictionarie with dtypes to use for upload.
    n_jobs :                Integer with number of jobs to use for parallelization.
    spark_mode :            String with mode to use when uploading to redshift with spark. Options are 'append', 'overwrite', 'ignore' and 'error'.
    """

    ## Set number of jobs

    if n_jobs == -1:
        n_jobs = multiprocessing.cpu_count()

    ## Upload data

    t_i = dt.datetime.now()
    n_try = 0
    succeeded = False
    while n_try < max_n_try and not succeeded:
        try:
            # Show shape of dataframe
            print(f"Shape of query dataframe -> {name} = {df.shape}")
            # Upload dataframe in parallel
            if not df.empty:
                # Parallelize only if dataframe is large enough (avoid locking up the system for small dataframes)
                if df.shape[0] / chunksize >= n_jobs:
                    print("Uploading chunked data in parallel...")
                    # Split dataframe in smaller chunks
                    df_split_iter = [
                        x for x in np.array_split(df, n_jobs) if not x.empty
                    ]
                    # Create iterators
                    table_name_iter = [table_name for x in df_split_iter]
                    schema_iter = [schema for x in df_split_iter]
                    mode_iter = [mode for x in df_split_iter]
                    conn_dict_iter = [conn_dict for x in df_split_iter]
                    custom_conn_str_iter = [custom_conn_str for x in df_split_iter]
                    connect_args_iter = [connect_args for x in df_split_iter]
                    chunksize_iter = [chunksize for x in df_split_iter]
                    method_iter = [method for x in df_split_iter]
                    dtypes_iter = [dtypes_dict for x in df_split_iter]
                    spark_mode_iter = [spark_mode for x in df_split_iter]
                    # Get results
                    parallel_results = parallel_execute(
                        parallel_to_sql,
                        df_split_iter,
                        table_name_iter,
                        schema_iter,
                        mode_iter,
                        conn_dict_iter,
                        custom_conn_str_iter,
                        connect_args_iter,
                        chunksize_iter,
                        method_iter,
                        dtypes_iter,
                        spark_mode_iter,
                    )

                    # Concatenate results
                    response_rows_affected = 0
                    for r in parallel_results:
                        response_rows_affected += r
                else:
                    print("Uploading whole data...")
                    response_rows_affected = parallel_to_sql(
                        df,
                        table_name,
                        schema,
                        mode,
                        conn_dict,
                        custom_conn_str,
                        connect_args,
                        chunksize,
                        method,
                        dtypes_dict,
                        spark_mode,
                    )
            else:
                # Set response rows affected to 0
                response_rows_affected = 0
            # Show rows affected
            print(f"Affected number of rows -> {name} = {response_rows_affected}")
            # Change status
            succeeded = True
        except Exception as e:
            # Set log path
            log_file_path = "logs"
            # Set log file name
            log_file_name = "upload_data"
            # Create logs folder
            os.makedirs(log_file_path, exist_ok=True)
            # Create logs
            mk_err_logs(
                log_file_path,
                log_file_name,
                sys._getframe().f_code.co_name + " -> " + name,
                traceback.format_exception(
                    sys.exc_info()[0], sys.exc_info()[1], sys.exc_info()[2]
                )[-1],
                mode="summary",
            )
            mk_err_logs(
                log_file_path,
                log_file_name,
                sys._getframe().f_code.co_name + " -> " + name,
                "".join(
                    traceback.format_exception(
                        sys.exc_info()[0], sys.exc_info()[1], sys.exc_info()[2]
                    )
                ),
                mode="detailed",
            )
            # Set response rows affected to 0
            response_rows_affected = 0
            # Change status
            succeeded = False
        # Update number of tries
        n_try += 1
    t_e = dt.datetime.now()
    print(f"Time elapsed in 'upload' process {name} -> {df.shape} = {t_e - t_i}")

    ## Create time execution logs

    # Set log path
    log_file_path = "logs"
    # Set log file name
    log_file_name = "upload_data_texec"
    # Create logs folder
    os.makedirs(log_file_path, exist_ok=True)
    # Create logs
    mk_texec_logs(
        log_file_path,
        log_file_name,
        sys._getframe().f_code.co_name + " -> " + name,
        t_e - t_i,
        obs=f"Shape of object = {df.shape}",
    )

    return response_rows_affected  # Response with number of rows affected


def sql_copy_data(
    s3_file_path,
    schema,
    table_name,
    conn_dict,
    access_key,
    secret_access_key,
    region,
    delimiter=",",
    header_row=1,
    type_format="csv",
    name=None,
    max_n_try=3,
):
    """
    Function to copy data to Redshift database from S3 bucket.

    s3_file_path:               String with S3 file paths to copy data from.
    schema:                     Schema to upload data to.
    table_name:                 Table name to upload data to.
    conn_dict:                  Dictionarie with server, database, uid and pwd information.
    access_key:                 String with access keys for S3 bucket.
    secret_access_key:          String with secret access keys for S3 bucket.
    region:                     String with regions for S3 bucket.
    delimiter:                  String with delimiter to use for copy command. Default is ','.
    header_row:                 Integer with header row to ignore. Default is 1.
    type_format:                String with type format to use for copy command. Default is 'csv'.
    name:                       Name to use for print statements.
    max_n_try:                  Integer with maximum number of tries to upload data.
    """

    ## Copy data

    t_i = dt.datetime.now()
    n_try = 0
    succeeded = False
    while n_try < max_n_try and not succeeded:
        try:
            # Create sql statement
            sql_stmt = f"COPY {schema}.{table_name} FROM '{s3_file_path}' ACCESS_KEY_ID '{access_key}' SECRET_ACCESS_KEY '{secret_access_key}' REGION '{region}' DELIMITER '{delimiter}' IGNOREHEADER {header_row} EMPTYASNULL FORMAT AS {type_format.upper()};"
            # Execute copy command
            response_rows_affected = sql_exec_stmt(sql_stmt, conn_dict, mode="redshift")
            # Show rows affected
            print(f"Affected number of rows -> {name} = {response_rows_affected}")
            # Change status
            succeeded = True
        except Exception as e:
            # Set log path
            log_file_path = "logs"
            # Set log file name
            log_file_name = "copy_data"
            # Create logs folder
            os.makedirs(log_file_path, exist_ok=True)
            # Create logs
            mk_err_logs(
                log_file_path,
                log_file_name,
                sys._getframe().f_code.co_name + " -> " + name,
                traceback.format_exception(
                    sys.exc_info()[0], sys.exc_info()[1], sys.exc_info()[2]
                )[-1],
                mode="summary",
            )
            mk_err_logs(
                log_file_path,
                log_file_name,
                sys._getframe().f_code.co_name + " -> " + name,
                "".join(
                    traceback.format_exception(
                        sys.exc_info()[0], sys.exc_info()[1], sys.exc_info()[2]
                    )
                ),
                mode="detailed",
            )
            # Set response rows affected to 0
            response_rows_affected = 0
            # Change status
            succeeded = False
        # Update number of tries
        n_try += 1
    t_e = dt.datetime.now()
    print(f"Time elapsed in 'copy' process {name} = {t_e - t_i}")

    ## Create time execution logs

    # Set log path
    log_file_path = "logs"
    # Set log file name
    log_file_name = "copy_data_texec"
    # Create logs folder
    os.makedirs(log_file_path, exist_ok=True)
    # Create logs
    mk_texec_logs(
        log_file_path,
        log_file_name,
        sys._getframe().f_code.co_name + " -> " + name,
        t_e - t_i,
        obs="Copy data from S3 bucket to database table",
    )

    return response_rows_affected  # Response with number of rows affected
